package subcluens

import data.Clustering._
import data.SubspaceCluster
import subcluens.datastructure.{DataTensor, TrueEntries}
import scala.util.Random

/**
  * Contains utility functions for generating data tensors and applying permutations of the ids
  */
object GenerateDataTensor {

  case class ShuffleMapping(backDim1: Map[Int, SubClusId], backDim2: Map[DpId, DpId], backDim3: Map[DimId, DimId])

  /**
    * The SubspaceCluster Ids do not have to be unique at this point, we simply don't care and use the sequence index instead
    *
    * @return the sorted data tensor;
    *         first dimension represents the subspaces with the sequence index
    *         second dimension is the dimension index
    *         third dimension is the datapoint index
    *
    */
  def generateSet(clusterings: Seq[SubspaceCluster]): (Set[(Int, DimId, DpId)], Map[SubspaceCluster, Int]) = {
    val clusterMapping = clusterings.zipWithIndex.toMap
    val data = (for {
      (sc, scIndex) <- clusterMapping
      dim <- sc.dimensions
      dp <- sc.dataPoints
    } yield {
      (scIndex, dim, dp)
    }).toSet

    (data, clusterMapping)
  }


  /**
    * Maps all ids of the data tensor to new ids using the randomizer parameter as a seed for the randomization.
    */
  def shuffleIds(dat: DataTensor[Int, DimId, DpId], randomizer: Int): (DataTensor[Int, DimId, DpId], ShuffleMapping) = {
    val rnd = new Random(randomizer)
    val trueEntries = dat.caCells(0)(0)(0).trueEntries
    val data = trueEntries.data
    val dim1Ids = rnd.shuffle(data.map(_._1).toSeq).zipWithIndex.toMap
    val dim2Ids = rnd.shuffle(data.map(_._2).toSeq).zipWithIndex.toMap
    val dim3Ids = rnd.shuffle(data.map(_._3).toSeq).zipWithIndex.toMap
    val dim1IdsBack = dim1Ids.map(x => x._2 -> x._1)
    val dim2IdsBack = dim2Ids.map(x => x._2 -> x._1)
    val dim3IdsBack = dim3Ids.map(x => x._2 -> x._1)

    val transformed = TrueEntries(trueEntries.data.map { case (r, c, t) => (dim1Ids(r), dim2Ids(c), dim3Ids(t)) })

    val transCats = for {d1 <- dat.caCells} yield {
      for {
        d2 <- d1} yield {
        for {
          cat <- d2
        } yield {
          cat.copy(trueEntries = transformed, cat.dim1Ids.map(dim1Ids), cat.dim2Ids.map(dim2Ids), cat.dim3Ids.map(dim3Ids))
        }
      }
    }
    DataTensor(transCats, dat.nrOfDim1El, dat.nrOfDim2El, dat.nrOfDim3El) -> ShuffleMapping(dim1IdsBack, dim2IdsBack, dim3IdsBack)
  }

  /**
    * Reverts the permutation of the ids.
    */
  def shuffleBackIds(dat: DataTensor[Int, DimId, DpId], mapping: ShuffleMapping) = {
    val trueEntries = dat.caCells(0)(0)(0).trueEntries
    val dim1IdsBack = mapping.backDim1
    val dim2IdsBack = mapping.backDim2
    val dim3IdsBack = mapping.backDim3

    val transformed = TrueEntries(trueEntries.data.map { case (r, c, t) => (dim1IdsBack(r), dim2IdsBack(c), dim3IdsBack(t)) })

    val transCats = for {d1 <- dat.caCells} yield {
      for {
        d2 <- d1} yield {
        for {
          cat <- d2
        } yield {
          cat.copy(trueEntries = transformed, cat.dim1Ids.map(dim1IdsBack), cat.dim2Ids.map(dim2IdsBack), cat.dim3Ids.map(dim3IdsBack))
        }
      }
    }
    DataTensor(transCats, dat.nrOfDim1El, dat.nrOfDim2El, dat.nrOfDim3El)
  }
}
