package subcluens.datastructure

import scala.collection.mutable


/**
  * This class represents a data tensor in different states of decomposition.
  *
  * @param caCells    the cross association cells
  * @param nrOfDim1El the number of elements in the first dimension
  * @param nrOfDim2El the number of elements in the second dimension
  * @param nrOfDim3El the number of elements in the third dimension
  * @tparam X the type of the first dimension
  * @tparam Y the type of the second dimension
  * @tparam Z the type of the third dimension
  */
case class DataTensor[X, Y, Z](caCells: IndexedSeq[IndexedSeq[IndexedSeq[CaCell[X, Y, Z]]]], nrOfDim1El: Int, nrOfDim2El: Int, nrOfDim3El: Int) {
  def getCaCells(idDim1: Int, idDim2: Int, idDim3: Int) = caCells(idDim1)(idDim2)(idDim3)

  def nrOf1DGroups: Int = caCells.length

  def nrOf2DGroups: Int = caCells(0).length

  def nrOf3DGroups: Int = caCells(0)(0).length

  def nrOfGroups = Seq(nrOf1DGroups, nrOf2DGroups, nrOf3DGroups)

  def dim1Indices = 0 until nrOf1DGroups

  def dim2Indices = 0 until nrOf2DGroups

  def dim3Indices = 0 until nrOf3DGroups


  @transient
  lazy val groupCountsDim1: IndexedSeq[Int] = new IndexedSeq[Int]() {
    override def length: Int = caCells.length

    override def apply(idx: Int): Int = caCells(idx)(0)(0).dim1Ids.size
  }

  @transient
  lazy val groupCountsDim2: IndexedSeq[Int] = new IndexedSeq[Int]() {
    override def length: Int = caCells(0).length

    override def apply(idx: Int): Int = caCells(0)(idx)(0).dim2Ids.size
  }
  @transient
  lazy val groupCountsDim3: IndexedSeq[Int] = new IndexedSeq[Int]() {
    override def length: Int = caCells(0)(0).length

    override def apply(idx: Int): Int = caCells(0)(0)(idx).dim3Ids.size
  }

  //We assume that each caCells share the same TrueEntries
  def getDim1Ids(groupNrDim1: Int): Set[X] = caCells(groupNrDim1)(0)(0).dim1Ids

  def getDim2Ids(groupNrDim2: Int): Set[Y] = caCells(0)(groupNrDim2)(0).dim2Ids

  def getDim3Ids(groupNrDim3: Int): Set[Z] = caCells(0)(0)(groupNrDim3).dim3Ids


  def getCaCellsInDim1(groupNrDim1: Int): TensorDim1Segment[X, Y, Z] = TensorDim1Segment(groupNrDim1, caCells)


  def getCaCellsInDim2(groupNrDim2: Int): TensorDim2Segment[X, Y, Z] = TensorDim2Segment(groupNrDim2, caCells)


  def getCaCellsInDim3(groupNrDim3: Int): TensorDim3Segment[X, Y, Z] = TensorDim3Segment(groupNrDim3, caCells)

  /**
    * Returns the total encoding costs of the data tensor
    */
  @transient
  lazy val encodingCosts: Double = EncodingCosts3d.determineDataTensorEncodingCosts(this)

  /**
    * Copies the properties of this data tensor with the new cross associations cells
    */
  def copyWithNewCaCells(caCells: IndexedSeq[IndexedSeq[IndexedSeq[CaCell[X, Y, Z]]]]) = new DataTensor(caCells, nrOfDim1El, nrOfDim2El, nrOfDim3El)

}

/**
  * Companion object which provides different factory functions for the DataTensor.
  */
object DataTensor {
  /**
    * We assume that the data set contains entries for all possible ids of X and Y
    * such that we can deduce the number of elements for each dimension based on the entries
    */
  def apply[X, Y, Z](data: Set[(X, Y, Z)]) = {
    val oneEntries = TrueEntries(data)
    val dim1Ids = data.map(_._1)
    val dim2Ids = data.map(_._2)
    val dim3Ids = data.map(_._3)
    val cadata = CaCell(oneEntries, dim1Ids, dim2Ids, dim3Ids)
    new DataTensor(Vector(Vector(Vector(cadata))), cadata.dim1Ids.size, cadata.dim2Ids.size, cadata.dim3Ids.size)
  }
}

/**
  * This class represents one cross-association cell of a data tensor.
  * the trueEntries are shared between all cells. The cell determines the number of true and false entries based on the
  * dimension-groups which are represented by the sets dim1Ids, dim2Ids, dim3Ids.
  */
case class CaCell[X, Y, Z](trueEntries: TrueEntries[X, Y, Z], dim1Ids: Set[X], dim2Ids: Set[Y], dim3Ids: Set[Z]) {
  private var cacheDim1 = collection.mutable.Map[X, Int]()
  private var cacheDim2 = collection.mutable.Map[Y, Int]()
  private var cacheDim3 = collection.mutable.Map[Z, Int]()


  val dim1Size = dim1Ids.size

  val dim2Size = dim2Ids.size

  val dim3Size = dim3Ids.size

  val size = dim1Size * dim2Size * dim3Size

  //Size of the matrix sizes if we take a slice of the respective dimension
  val matrixSizeForDim3Idx = dim1Ids.size * dim2Ids.size
  val matrixSizeForDim2Idx = dim1Ids.size * dim3Ids.size
  val matrixSizeForDim1Idx = dim2Ids.size * dim3Ids.size

  //This is the only operation in this class so we can make use of the caches and use a var
  //so we can reuse the computation if needed e.g. when we assign a different clusterGroup flag
  private var _trueCount: Option[Int] = None

  def trueCount: Int = {
    if (_trueCount.isEmpty) {
      if (cacheDim1.nonEmpty) {
        _trueCount = Some(dim1Ids.foldLeft(0)((sum, x) => trueCountForDim1Idx(x) + sum))
      } else if (cacheDim2.nonEmpty) {
        _trueCount = Some(dim2Ids.foldLeft(0)((sum, x) => trueCountForDim2Idx(x) + sum))
      } else {
        _trueCount = Some(dim3Ids.foldLeft(0)((sum, x) => trueCountForDim3Idx(x) + sum))
      }
    }
    val tc = _trueCount.get
    tc
  }


  /**
    *
    * Determines the number of true entries in the matrix slice of the given row element
    */
  def trueCountForDim1Idx(idx: X) = {
    assert(dim1Ids.contains(idx))
    this.synchronized {
      cacheDim1.getOrElseUpdate(idx, {
        val trueEntriesForIdx = trueEntries.givenDim1Id(idx)
        if (trueEntriesForIdx.size < matrixSizeForDim1Idx) {
          var counter = 0
          for {
            (dim2Idx, dim3Idx) <- trueEntriesForIdx
          } {
            if (dim2Ids.contains(dim2Idx) && dim3Ids.contains(dim3Idx)) {
              counter += 1
            }
          }
          counter
        } else {
          var counter = 0
          for {
            dim2Idx <- dim2Ids
            dim3Idx <- dim3Ids
          } {
            if (trueEntriesForIdx.contains((dim2Idx, dim3Idx))) {
              counter += 1
            }
          }
          counter
        }
      })
    }
  }


  /**
    *
    * Determines the number of true entries in the matrix slice of the given column element
    */
  def trueCountForDim2Idx(idx: Y) = {
    assert(dim2Ids.contains(idx))
    this.synchronized {
      cacheDim2.getOrElseUpdate(idx, {
        val trueEntriesForIdx = trueEntries.givenDim2Id(idx)
        if (trueEntriesForIdx.size < matrixSizeForDim2Idx) {
          var counter = 0
          for {
            (dim1Idx, dim3Idx) <- trueEntriesForIdx
          } {
            if (dim1Ids.contains(dim1Idx) && dim3Ids.contains(dim3Idx)) {
              counter += 1
            }
          }
          counter
        } else {
          var counter = 0
          for {
            dim1Idx <- dim1Ids
            dim3Idx <- dim3Ids
          } {
            if (trueEntriesForIdx.contains((dim1Idx, dim3Idx))) {
              counter += 1
            }
          }
          counter
        }
      })
    }
  }

  /**
    *
    * Determines the number of true entries in the matrix slice of the given tube element
    */
  def trueCountForDim3Idx(idx: Z) = {
    assert(dim3Ids.contains(idx))
    this.synchronized {
      cacheDim3.getOrElseUpdate(idx, {
        var counter = 0
        val trueEntriesForIdx = trueEntries.givenDim3Id(idx)
        if (trueEntriesForIdx.size < matrixSizeForDim3Idx) {
          for {
            (dim1Idx, dim2Idx) <- trueEntriesForIdx
          } {
            if (dim1Ids.contains(dim1Idx) && dim2Ids.contains(dim2Idx)) {
              counter += 1
            }
          }
          counter
        } else {
          var counter = 0
          for {
            dim1Idx <- dim1Ids
            dim2Idx <- dim2Ids
          } {
            if (trueEntriesForIdx.contains((dim1Idx, dim2Idx))) {
              counter += 1
            }
          }
          counter
        }
      })
    }
  }

  def trueRateForDim3Idx(idx: Z) = trueCountForDim3Idx(idx).toDouble / matrixSizeForDim3Idx

  def trueRateForDim2Idx(idx: Y) = trueCountForDim2Idx(idx).toDouble / matrixSizeForDim2Idx

  def trueRateForDim1Idx(idx: X) = trueCountForDim1Idx(idx).toDouble / matrixSizeForDim1Idx

  //Term 5 nrOf oneEntries
  lazy val encodingCostNrOfTrues = EncodingCosts3d.caCellNrOfTruesCosts(size)

  lazy val (encodingCostsTrue, encodingCostsFalse) = EncodingCosts3d.determinePerEventEncodingCosts(nrOfTrues = trueCount.toDouble, nrOfEntries = size)

  //We decided to use a fixed lower bound on the probability for the regrouping
  lazy val (encodingCostsTrueWithMinProb, encodingCostsFalseWithMinProb) = EncodingCosts3d.determinePerEventEncodingCosts(nrOfTrues = trueCount.toDouble, nrOfEntries = size, minProbability = 1.0e-4)

  lazy val dataEncodingCosts = EncodingCosts3d.determineEncodingCosts(nrOfTrues = trueCount, nrOfFalse = size - trueCount, trueEncoding = encodingCostsTrue, falseEncoding = encodingCostsFalse)

  lazy val trueRate = trueCount / size.toDouble

  /**
    * Specialized copy function which reuses the cache
    */
  def copyWithNewDim1(dim1Ids: Set[X]) = {
    val n = CaCell(trueEntries, dim1Ids, dim2Ids, dim3Ids)
    n.cacheDim1 = this.cacheDim1.filter(x => dim1Ids(x._1))
    n
  }

  /**
    * Specialized copy function which reuses the cache
    */
  def copyWithNewDim2(dim2Ids: Set[Y]) = {
    val n = CaCell(trueEntries, dim1Ids, dim2Ids, dim3Ids)
    n.cacheDim2 = this.cacheDim2.filter(x => dim2Ids(x._1))
    n
  }

  /**
    * Specialized copy function which reuses the cache
    */
  def copyWithNewDim3(dim3Ids: Set[Z]) = {
    val n = CaCell(trueEntries, dim1Ids, dim2Ids, dim3Ids)
    n.cacheDim3 = this.cacheDim3.filter(x => dim3Ids(x._1))
    n
  }
}


/**
  * A simple wrapper class for the set of triplets with some utility functions.
  */
case class TrueEntries[X, Y, Z](data: Set[(X, Y, Z)]) {
  val (dim1Map, dim2Map, dim3Map) = {
    val dim1Map = mutable.Map[X, mutable.Set[(Y, Z)]]()
    val dim2Map = mutable.Map[Y, mutable.Set[(X, Z)]]()
    val dim3Map = mutable.Map[Z, mutable.Set[(X, Y)]]()
    for {
      (x, y, z) <- data
    } {
      val dim1IdsSet = dim1Map.getOrElseUpdate(x, mutable.Set[(Y, Z)]())
      dim1IdsSet += y -> z
      val dim2IdsSet = dim2Map.getOrElseUpdate(y, mutable.Set[(X, Z)]())
      dim2IdsSet += x -> z
      val dim3IdsSet = dim3Map.getOrElseUpdate(z, mutable.Set[(X, Y)]())
      dim3IdsSet += x -> y
    }
    (dim1Map.map(x => x._1 -> x._2.toSet).toMap, dim2Map.map(x => x._1 -> x._2.toSet).toMap, dim3Map.map(z => z._1 -> z._2.toSet).toMap)
  }

  def givenDim1Id(id: X): Set[(Y, Z)] = dim1Map.getOrElse(id, Set.empty)

  def givenDim2Id(id: Y): Set[(X, Z)] = dim2Map.getOrElse(id, Set.empty)

  def givenDim3Id(id: Z): Set[(X, Y)] = dim3Map.getOrElse(id, Set.empty)
}
