package subcluens.datastructure

import _root_.utils.LogarithmUtils._

import scala.math._

/**
  * A utility object which contains all aspects of the encoding costs determination for a data tensor.
  * We extracted it to play with different combinations
  */
object EncodingCosts3d {

  //Determines term 5 based on the number of entries
  @inline
  def caCellNrOfTruesCosts(nrOfEntries: Double): Double = ceil(log2(nrOfEntries + 1))


  /**
    * Determines the encoding costs for an element with the specified amount of true entries
    *
    * @param sliceSize determines how many elements a entry has in this matrix (depends on the current dimension)
    */
  @inline
  def determinesEncodingCostsOfElement[X, Y, Z](nrOfTrues: Int, cat: CaCell[X, Y, Z], sliceSize: CaCell[X, Y, Z] => Int): Double = {
    val nrOfFalse = sliceSize(cat) - nrOfTrues
    determineEncodingCosts(nrOfTrues = nrOfTrues, nrOfFalse = nrOfFalse,
      trueEncoding = cat.encodingCostsTrueWithMinProb, falseEncoding = cat.encodingCostsFalseWithMinProb)
  }


  /**
    *  Determines term 6 of the encoding costs  for one entry, but it has an additional parameter which defines a lower bound for the probability.
    *  This is used during regrouping
    */
  @inline
  def determinePerEventEncodingCosts(nrOfTrues: Double, nrOfEntries: Double, minProbability: Double): (Double, Double) = {
    val trueProbability = min(max(minProbability, nrOfTrues / nrOfEntries), 1 - minProbability)
    val falseProbability = 1 - trueProbability
    val trueEncodingCosts = if (trueProbability > 0d) log2(1d / trueProbability) else Double.PositiveInfinity
    val falseEncodingCosts = if (falseProbability > 0d) log2(1d / falseProbability) else Double.PositiveInfinity
    (trueEncodingCosts, falseEncodingCosts)
  }


  /**
    * Determines term 6 of the encoding costs for one entry
    */
  @inline
  def determinePerEventEncodingCosts(nrOfTrues: Double, nrOfEntries: Double): (Double, Double) = {
    val trueProbability = nrOfTrues / nrOfEntries
    val falseProbability = 1 - trueProbability
    val trueEncodingCosts = if (trueProbability > 0d) log2(1d / trueProbability) else Double.PositiveInfinity
    val falseEncodingCosts = if (falseProbability > 0d) log2(1d / falseProbability) else Double.PositiveInfinity
    (trueEncodingCosts, falseEncodingCosts)
  }

  /**
    * Determines term 6 of the encoding costs for the all entries
    */
  @inline
  def determineEncodingCosts(nrOfTrues: Double, nrOfFalse: Double, trueEncoding: Double, falseEncoding: Double): Double = {
    try {
      if (nrOfTrues == 0 && nrOfFalse == 0) {
        0d //This happens e.g. while regrouping
      } else {
        if (nrOfTrues == 0) {
          0d
        } else if (nrOfFalse == 0d) {
          0d
        } else {
          nrOfTrues * trueEncoding + nrOfFalse * falseEncoding
        }
      }
    } catch {
      case e: Exception =>
        println(s"$nrOfTrues, $nrOfFalse, $trueEncoding, $falseEncoding")
        throw e
    }
  }

  /**
    * Determines term 4 of the description encoding.
    *
    */
  //-1 because the last groupCount can be determined based on the totalNrOfObjects-Transferred object
  @inline
  def calculateEncodingCostForNrOfEntriesPerGroup(groupCounts: Seq[Int], totalNrOfObjects: Int): Double = (groupCounts.length - 1).toDouble * ceil(log2(totalNrOfObjects))


  /**
    * Determines the encoding costs based on the given input
    * with entity we mean the nr of rows or cols (which ever the user currently tries to determine)
    */
  @inline
  def determineAvgEncodingCostsPerEntity(nrOfTrues: Double, matrixSize: Double, nrOfEntities: Double): Double = {
    val nrOfFalse = matrixSize - nrOfTrues
    val (trueEncodingCosts, falseEncodingCosts) = determinePerEventEncodingCosts(nrOfTrues = nrOfTrues, nrOfEntries = matrixSize)
    determineEncodingCosts(nrOfTrues = nrOfTrues, nrOfFalse = nrOfFalse, trueEncoding = trueEncodingCosts, falseEncoding = falseEncodingCosts) / nrOfEntities
  }

  /**
    * Determines the total encoding costs for a data tensor
    */
  @inline
  def determineDataTensorEncodingCosts(dt: DataTensor[_, _, _]) = {
    var costOfActualData = 0d
    for {
      tensorSeqSeq <- dt.caCells
      tensorSeq <- tensorSeqSeq
      cat <- tensorSeq
    } {
      costOfActualData += cat.dataEncodingCosts + cat.encodingCostNrOfTrues
    }

    //Term 3 send the number of groups
    val costOfD1NrOfGroups = logStar2(dt.nrOf1DGroups)
    val costOfD2NrOfGroups = logStar2(dt.nrOf2DGroups)
    val costOfD3NrOfGroups = logStar2(dt.nrOf3DGroups)

    //Term 4 send the number of rows in each row group and the number of columns in each column group
    val costOfD1GroupCount = calculateEncodingCostForNrOfEntriesPerGroup(dt.groupCountsDim1, dt.nrOfDim1El)
    val costOfD2GroupCount = calculateEncodingCostForNrOfEntriesPerGroup(dt.groupCountsDim2, dt.nrOfDim2El)
    val costOfD3GroupCount = calculateEncodingCostForNrOfEntriesPerGroup(dt.groupCountsDim3, dt.nrOfDim3El)

    val totalCost = costOfD1NrOfGroups + costOfD2NrOfGroups + costOfD3NrOfGroups +
      costOfD1GroupCount + costOfD2GroupCount + costOfD3GroupCount +
      costOfActualData
    totalCost
  }
}
