package subcluens.operations

import subcluens.datastructure.{CaCell, EncodingCosts3d, TensorSegment}

import scala.collection.mutable

/**
  * The trait represents the regrouping functionality which can be abstracted away from the specific dimension through the TensorSegments
  */
trait AbstractRegroup {
  type mMap[A] = mutable.Map[Int, mutable.Set[A]]

  /**
    * Determines the index of the group with the least encoding costs for the given element
    *
    * @param segEntryIdx   the entry we seek the best new segment for
    * @param tensorSegment the current segment the segEntryIdx belongs to
    * @param allSegments   all segments in which we seek a new "home" for segEntryIdx
    * @return
    */
  protected def findBestEncodingGroupForData[A, X, Y, Z](segEntryIdx: A,
                                                         tensorSegment: TensorSegment[A, X, Y, Z],
                                                         allSegments: IndexedSeq[TensorSegment[A, X, Y, Z]]): Int = {
    var bestGroupIdx = -1
    var bestEncodingCosts = Double.PositiveInfinity
    for {
      modGroupIdx <- allSegments.indices
    } {
      var currentEncodingCosts = 0d
      for {
        groupDimAIdx <- tensorSegment.dimAIndices
        groupDimBIdx <- tensorSegment.dimBIndices
      } {
        val cat = allSegments(modGroupIdx).caCell(groupDimAIdx, groupDimBIdx)
        val trueCount = tensorSegment.getTrueCountForEntry(groupDimAIdx, groupDimBIdx)(segEntryIdx)
        //The only place where we use the minProbability
        currentEncodingCosts +=
          EncodingCosts3d.determinesEncodingCostsOfElement[X, Y, Z](trueCount, cat, determineEntryMatrixSize)
      }
      if (currentEncodingCosts < bestEncodingCosts) {
        bestEncodingCosts = currentEncodingCosts
        bestGroupIdx = modGroupIdx
      }
    }
    bestGroupIdx
  }

  /**
    * This function determines a mapping for the ids which should be moved to a different group.
    * @return None if nothing todo A Map with the changes, if the index is not contained in the map this group is empty.
    *         If it is without update its None and if there are updates its some
    */
  protected def regroupIds[A, X, Y, Z](segments: IndexedSeq[TensorSegment[A, X, Y, Z]]): Option[Map[Int, Option[Set[A]]]] = {
    val removeSets = mutable.Map[Int, mutable.Set[A]]()
    val addSets = mutable.Map[Int, mutable.Set[A]]()

    for {
      segmentGroupIdx <- segments.indices
    } {
      val rowGroupSlice = segments(segmentGroupIdx)
      for {
        entryIdx <- rowGroupSlice.groupIds
      } {
        val newSegmentGroupIdx = findBestEncodingGroupForData(entryIdx, rowGroupSlice, segments)
        if (newSegmentGroupIdx != segmentGroupIdx) {
          //Another row-group encodes this line-element better
          val remSet = removeSets.getOrElseUpdate(segmentGroupIdx, mutable.Set())
          remSet += entryIdx
          val addSet = addSets.getOrElseUpdate(newSegmentGroupIdx, mutable.Set())
          addSet += entryIdx
        }
      }
    }

    if (removeSets.isEmpty && addSets.isEmpty) {
      None
    } else {
      val idUpdates = segments.indices.flatMap { segmentGroupIdx =>
        val currentIds = segments(segmentGroupIdx).groupIds
        val removeEntries = removeSets.get(segmentGroupIdx)
        val addEntries = addSets.get(segmentGroupIdx)
        if (removeEntries.isEmpty && addEntries.isEmpty) {
          Some(segmentGroupIdx -> None)
        } else {
          val newIds = currentIds -- removeEntries.getOrElse(Set.empty) ++ addEntries.getOrElse(Set.empty)
          if (newIds.isEmpty)
            None
          else {
            Some(segmentGroupIdx -> Some(newIds))
          }
        }
      }.toMap
      Some(idUpdates)
    }
  }

  //Determines CurrentTensorProps(nrOfMatrixElements: Int, //Fixed as only the number of row elements changes
  protected def determineEntryMatrixSize[X, Y, Z](caTensor: CaCell[X, Y, Z]): Int
}